import java.util.*;

/**
 * Implements a ball as a PhysicsElement
 */
public class Ball extends PhysicsElement {
    private final float mass;
    private final float radius;
    private double crossSection;				// to compute air friction 
    private Vector2D pos_t;						// current position at time t
    private Vector2D pos_tPlusDelta;				// next position at t + delta
    private Vector2D speed_t;						// speed at time t
    private Vector2D speed_tPlusDelta;			// speed at t + delta
    private Vector2D a_t;							// acceleration at time t
    private Vector2D a_tMinusDelta;				// acceleration at t-delta

	/**
	 * Ball constructor 
	 */
    public Ball(float mass, float r, Vector2D position, Vector2D speed, MyWorld world) {
        super();
        this.mass = mass;
        radius = r;
        pos_t = position;
        speed_t = speed;
        a_t = new Vector2D (0, 0);
        a_tMinusDelta = getNetForce(world.getGravity(), world.getViscosity()).scale(1/mass);
        crossSection = r;
        //-------------------------
        world.addElement(this);
    }

	/**
	 * Get the mass of the ball
	 */
    @Override
    public float getMass() {
        return mass;
    }
    
    /**
	 * Get the radius of the ball
	 */
    @Override
    public float getRadius() {
        return radius;
    }

	/**
	 * Get the current position of the ball
	 */
    @Override
    public Vector2D getPosition() {
        return pos_t;
    }
    
	/**
	 * Get the current speed of the ball. It returns a vector.
	 * Strictly this should be named velocity. 
	 */
    @Override
    public Vector2D getSpeed() {
        return speed_t;
    }

	/**
	 * Analyzes if this object is colliding with another PhysicsElement.
	 * If the PhysicsElement is a wall, then we ask the wall if this
	 * ball is colliding with it.
	 */
    @Override
    public boolean collide(PhysicsElement pe) {
        if (pe instanceof Wall) {
            Wall w = (Wall) pe;
            return w.collide(this);
        } else {
            float distance;
            distance = pos_t.getSubtraction (pe.getPosition ()).norm();
            return distance < (pe.getRadius()+radius);
        }
    }

	/**
	 * Returns the current net force being applied to the ball.
	 */
    private Vector2D getNetForce(Vector2D g, float viscosity) {
        Vector2D gravityForce;
        Vector2D frictionForce;
        Vector2D netForce;

        gravityForce = g.getScaled (mass);
        frictionForce = speed_t.getNegative().scale ( (float) (viscosity*crossSection));
        netForce = gravityForce.getSum(frictionForce);

        return(netForce);	 
    }

	/**
	 * Compute the next state of the ball.
	 * This means it calculates the next position,
	 * speed and acceleration of the ball, without
	 * actually modifying this variables.
	 */
    @Override
    public void computeNextState(double delta_t, MyWorld w) {
        Vector2D g = w.getGravity ();
        float viscosity = w.getViscosity ();
        
        Vector2D speed;			//temporal variable used to store the calculated value of the speed after a collision
        float collDirSpeed;     //componente de la velocidad en dirección de la colicion
        float collPerSpeed;     //componente de la velocidad en dirección perpendicular a la colicion
        float nearECollDS;      //near element collision direction speed // la velocidad del otro elemendo en la direccion de colision
        Vector2D collisionDir;  //dirección en la que ocurre la colicion
        Vector2D collPerpDir;   //direccion perpendicular a la colocion
        float nearEMass;		//mass of the colliding object
        PhysicsElement nearE = w.getNear (this);	//variable to hold the colliding object

        if (nearE != null) { /* elastic collision */
        
        	//if the colliding object is a wall then only invert one of the component of the velocity
            if (nearE instanceof Wall){
                Wall wl = (Wall) nearE;
                if (wl.getAxis() == 'x') {
                    collisionDir = new Vector2D(1,0);
                    collPerpDir = collisionDir.getPerpendicular();
                } else {
                    collisionDir = new Vector2D(0,1);
                    collPerpDir = collisionDir.getPerpendicular();
                }
                collDirSpeed = speed_t.getProjectionMagnitude (collisionDir);
                collPerSpeed = speed_t.getProjectionMagnitude (collPerpDir);
                
                collDirSpeed = -collDirSpeed;
                
                //recompone el vector velocidad
                speed = collisionDir.scale(collDirSpeed);
                speed.sum(collPerpDir.scale(collPerSpeed));
                
            //if the colliding object is not a wall:
            } else {
                nearEMass = nearE.getMass ();

                //Obtiene vectores unitarios en direccion y perpendicular a la colision
                collisionDir = pos_t.getSubtraction (nearE.getPosition ()).getUnitVector();
                collPerpDir = collisionDir.getPerpendicular();

                //descompone el vector velocidad en vectores del nuevo eje de coordenadas
                collDirSpeed = speed_t.getProjectionMagnitude (collisionDir);
                collPerSpeed = speed_t.getProjectionMagnitude (collPerpDir);
                nearECollDS  = nearE.getSpeed().getProjectionMagnitude(collisionDir);

                //calcula la velocidad en direccion de la colision, despues de la colision
                collDirSpeed = (collDirSpeed*(mass - nearEMass) + 2*nearEMass*nearECollDS)/(mass + nearEMass);

                //recompone el vector velocidad
                speed = collisionDir.scale(collDirSpeed);
                speed.sum(collPerpDir.scale(collPerSpeed));

            }
            a_tMinusDelta = getNetForce(g, viscosity).scale(mass);
            
        } else 
            speed = speed_t;
		
		//calculates the next acceleration, velocity and position.
        a_t = getNetForce(g, viscosity).scale(mass);
        speed_tPlusDelta = speed.getSum(a_t.getScaled(3).subtact(a_tMinusDelta).scale(delta_t/2));
        pos_tPlusDelta = a_t.getScaled(4).subtact(a_tMinusDelta).scale(delta_t*delta_t/6).sum(pos_t).sum(speed.getScaled(delta_t));
    }

	/**
	 * Updates the position, speed and acceleration with the next values.
	 */
    @Override
    public void updateState() {
        pos_t = pos_tPlusDelta;
        speed_t = speed_tPlusDelta;
        a_tMinusDelta = a_t;
    }

	/**
	 * Returns a desctiption of the ball.
	 */
    @Override
    public String getDescription() {
        return "Ball_" + getId() + ": x ";
    }

	/**
	 * Returns a string with the actual position of the ball.
	 */
    @Override
    public String getState() {
        return String.format ("%.6f, %.6f", pos_t.getX(), pos_t.getY());
    }
}
