import java.io.*;
import java.util.*;

/**
 * Simulates a world that contains basic physic elements.
 */
public class MyWorld {
	private float viscosity;
	private Vector2D g;			//Gravity
	private PrintStream out;
	private Vector<PhysicsElement> elements; // Vector can grow, arrays cannot.

	/**
	 * Constructor.
	 */
	public MyWorld(PrintStream output) {
		g = new Vector2D(0, 0);
		viscosity = 0.0f;
		out = output;
		elements = new Vector<PhysicsElement>();
	}
	
	/**
	 * Set the gravity of this world.
	 */
	public void setGravity (Vector2D gravity) {
		g = gravity;
	}
	
	/**
	 * Return the gravity of this world.
	 */
	public Vector2D getGravity () {
		return g;
	}
	
	/**
	 * Set the viscosity of this world.
	 */
	public void setViscosity(float v){
		viscosity = v;
	}
	
	/**
	 * Return the viscosity of this world.
	 */
	public float getViscosity(){
		return viscosity;
	}
	
	/**
	 * Add a Physics Element to this world.
	 */
	public void addElement(PhysicsElement e) {
		elements.add(e);
	}
	
	/**
	 * Test if an element is colliding with the element passed as parameter.
	 */
	public PhysicsElement getNear(PhysicsElement anchor) {
		for (PhysicsElement e: elements) 
			if (anchor.collide(e) && (anchor!=e))
				return e;
		return null;		
	}
	
	/**
	 * Print the headers of the final data table.
	 */
	private void printSystemHeaders() {
		out.print("Time");
		for (PhysicsElement e: elements)  // print header
			out.print(",\t\t"+e.getDescription());
		out.println();
	}  
	
	/**
	 * Print the state of all elements for a time t.
	 */
	private void printSystemState(double t) {
		out.print (String.format ("%.2f", t));
		for (PhysicsElement e: elements)  
			out.print(",\t\t" + e.getState() );  // print each element state
		out.println();
	}		  

	/**
	 * Simulation routine. Increases the time, calculates the state of all elements and print the resoults.
	 */
	public void simulate ( float delta_t, float endTime, float samplingTime){
		printSystemHeaders();
		for (double t=0, nextStateTime=0; t<endTime; t+=delta_t){
			if (t >= nextStateTime) {		 // decide if it is time to print state 
				printSystemState(t);
				nextStateTime += samplingTime;
			} 
			for (PhysicsElement e: elements)  // for each element
				e.computeNextState(delta_t, this); // compute next state
			for (PhysicsElement e: elements)  // for all elements
				e.updateState();					 // update its state
		}
	}
} 
